package cashmachine;

import static org.hamcrest.Matchers.is;
import static org.junit.Assert.*;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.*;

import java.util.Arrays;

import org.junit.Test;

import cashmachine.api.AccountService;
import cashmachine.api.CashMachine;
import cashmachine.api.CashMachineException;
import cashmachine.api.MachineConnector;
import cashmachine.impl.CashMachineImpl;
import cashmachine.model.Account;

public class CashMachineTest
{
	private static final int AC1_ID = 1;
	private static final int AC1_PIN = 10;
	private static final int AC1_BALANCE = 10000;
	private static final int AC1_LIMIT = 1000;
	private static final int AC2_ID = 2;
	private static final int AC2_PIN = 20;
	private static final int AC2_BALANCE = 22;
	private static final int AC2_LIMIT = 1000;
	private static final int INVALID_ID = 3;
	private static final int DEFAULT_FILL_10 = 100;
	private static final int DEFAULT_FILL_20 = 50;
	private static final int DEFAULT_FILL_50 = 20;
	private static final int DEFAULT_FILL_100 = 10;
	private static final int DEFAULT_FILL_200 = 5;

	private Account createAccount1() {
		return new Account(AC1_ID, AC1_PIN, AC1_BALANCE, AC1_LIMIT, Arrays.asList(AC2_ID));
	}

	private Account createAccount2() {
		return new Account(AC2_ID, AC2_PIN, AC2_BALANCE, AC2_LIMIT, Arrays.<Integer> asList());
	}

	private MachineConnector createMockedConnector() {
		return mock(MachineConnector.class);
	}

	private AccountService createMockedAccountService() throws CashMachineException {
		AccountService a = mock(AccountService.class);
		when(a.isValidAccountId(AC1_ID)).thenReturn(true);
		when(a.validatePin(AC1_ID, AC1_PIN)).thenReturn(createAccount1());
		when(a.getWithdrawalAccount(AC1_ID, AC1_PIN, AC1_ID)).thenReturn(createAccount1());
		when(a.getWithdrawalAccount(AC1_ID, AC1_PIN, AC2_ID)).thenReturn(createAccount2());
		when(a.withdraw(eq(AC1_ID), eq(AC1_PIN), eq(AC1_PIN), anyInt())).thenReturn(1);
		when(a.withdraw(eq(AC1_ID), eq(AC1_PIN), eq(AC2_PIN), anyInt())).thenReturn(1);
		return a;
	}

	private CashMachine createCashMachine() throws CashMachineException {
		return createCashMachine(createMockedConnector());
	}

	private CashMachine createCashMachine(MachineConnector machineConnector)
	throws CashMachineException {
		return createCashMachine(machineConnector, createMockedAccountService());
	}

	private CashMachine createCashMachine(AccountService accountService) {
		return createCashMachine(createMockedConnector(), accountService);
	}

	private CashMachine createCashMachine(MachineConnector connector, AccountService accountService) {
		return createCashMachine(connector, accountService, DEFAULT_FILL_10,
				DEFAULT_FILL_20, DEFAULT_FILL_50, DEFAULT_FILL_100, DEFAULT_FILL_200);
	}

	private CashMachine createCashMachine(MachineConnector machineConnector,
			AccountService accountService, int amount10, int amount20, int amount50, int amount100,
			int amount200) {
		return new CashMachineImpl(machineConnector, accountService, amount10, amount20, amount50,
				amount100, amount200);
	}

	@Test
	public void fillLevels_shouldReturnRightAmount() throws CashMachineException {
		CashMachine cm = createCashMachine();
		assertThat(cm.getTenFillLevel(), is(DEFAULT_FILL_10));
		assertThat(cm.getTwentyFillLevel(), is(DEFAULT_FILL_20));
		assertThat(cm.getFiftyFillLevel(), is(DEFAULT_FILL_50));
		assertThat(cm.getOneHundredFillLevel(), is(DEFAULT_FILL_100));
		assertThat(cm.getTwoHundredFillLevel(), is(DEFAULT_FILL_200));
	}
	
	/*--------------------------------------------------------------------------*/
	
	// TODO put your tests here
	@Test
	public void setAccountId_shouldBeValid()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertTrue(cm.setAccountId(AC1_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void setAccountId_shouldBeInvalid()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertFalse(cm.setAccountId(INVALID_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void setAccountId_shouldThrow()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertTrue(cm.setAccountId(AC1_ID));
		cm.setAccountId(AC1_ID);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void setPin_shouldBeValid()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertTrue(cm.setAccountId(AC1_ID));
		assertTrue(cm.setPin(AC1_PIN));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void setPin_shouldBeInvalid()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertTrue(cm.setAccountId(AC1_ID));
		assertFalse(cm.setPin(AC1_PIN + 1));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void setPin_shouldThrow()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		cm.setPin(AC1_PIN);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void isCardValid_shouldBeValid()
	throws CashMachineException
	{
		CashMachine cm = createCashMachine();
		assertTrue(cm.setAccountId(AC1_ID));
		assertTrue(cm.isCardValid());
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void dispenseAmount_should_open_moneyTray()
	throws CashMachineException
	{
		MachineConnector mc = createMockedConnector();
		CashMachine cm = createCashMachine(mc);
		assertFalse(cm.isMoneyTrayOpen());
		assertTrue(cm.setAccountId(AC1_ID));
		assertTrue(cm.setPin(AC1_PIN));
		java.util.LinkedList<Integer> list = new java.util.LinkedList<Integer>(cm.getWithdrawalAccountIds());
		assertFalse(list.isEmpty());
		cm.setWithdrawalAccountId(list.get(0));
		assertTrue(cm.withdraw((AC1_LIMIT > 10) ? 10 : 0));
		assertTrue(cm.dispenseAmount());
		assertTrue(cm.isMoneyTrayOpen());
		verify(mc).openMoneyTray();
	}
	
	/*--------------------------------------------------------------------------*/
	
	public class DispenseMoneyTester
	{
		private int ten, twenty, fifty, onehundret, twohundret;

		public DispenseMoneyTester(int ten, int twenty, int fifty, int onehundret, int twohundret)
		{
			this.ten = ten;
			this.twenty = twenty;
			this.fifty = fifty;
			this.onehundret = onehundret;
			this.twohundret = twohundret;
		}
		
		public int getTenFillLevel()
		{
			return ten;
		}
		
		public int getTwentyFillLevel()
		{
			return twenty;
		}
		
		public int getFiftyFillLevel()
		{
			return fifty;
		}
		
		public int getOneHundredFillLevel()
		{
			return onehundret;
		}
		
		public int getTwoHundredFillLevel()
		{
			return twohundret;
		}

		public boolean dispense(int amount)
		{
			int tmpten = 0, tmptwenty = 0, tmpfifty = 0;
			int tmponehundret = 0, tmptwohundret = 0;
			
			while(amount > 500)
			{
				tmptwohundret += 2;
				amount -= (2 * 200);
			}
			while (amount <= 500 && amount > 200)
			{
				tmponehundret += 1;
				tmptwenty += 4;
				amount -= (1 * 100 + 4 * 20);
			}
			while(amount <= 200 && amount > 110)
			{
				tmpfifty += 1;
				tmptwenty += 3;
				amount -= (1 * 50 + 3 * 20);
			}
			while(amount <= 110 && amount > 50)
			{
				tmpfifty += 1;
				amount -= (1 * 50);
			}
			while (amount > 0)
			{
				tmpten += 1;
				amount -= (1 * 10);
			}
			
			if (amount != 0 || tmpten > ten || tmptwenty > twenty || tmpfifty > fifty
					|| tmponehundret > onehundret || tmptwohundret > twohundret)
				return false;
			
			ten -= tmpten;
			twenty -= tmptwenty;
			fifty -= tmpfifty;
			onehundret -= tmponehundret;
			twohundret -= tmptwohundret;
			return true;
		}
	}
	
	/*--------------------------------------------------------------------------*/

	@Test
	public void dispenseMoney()
	throws CashMachineException
	{
		int money = 660;
		CashMachine cm = createCashMachine();
		DispenseMoneyTester tester = new DispenseMoneyTester(cm.getTenFillLevel(), cm.getTwentyFillLevel(),
				cm.getFiftyFillLevel(), cm.getOneHundredFillLevel(), cm.getTwoHundredFillLevel());
		assertTrue(cm.setAccountId(AC1_ID));
		assertTrue(cm.setPin(AC1_PIN));
		java.util.LinkedList<Integer> list = new java.util.LinkedList<Integer>(cm.getWithdrawalAccountIds());
		assertFalse(list.isEmpty());
		cm.setWithdrawalAccountId(list.get(0));
		assertTrue(cm.withdraw(money));
		assertTrue(cm.dispenseAmount());
		assertTrue(tester.dispense(money));
		assertEquals(tester.getTenFillLevel(),        cm.getTenFillLevel());
		assertEquals(tester.getTwentyFillLevel(),     cm.getTwentyFillLevel());
		assertEquals(tester.getFiftyFillLevel(),      cm.getFiftyFillLevel());
		assertEquals(tester.getOneHundredFillLevel(), cm.getOneHundredFillLevel());
		assertEquals(tester.getTwoHundredFillLevel(), cm.getTwoHundredFillLevel());
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void setPin_shouldBanCard()
	throws CashMachineException
	{
		AccountService as = createMockedAccountService();
		CashMachine cm = createCashMachine(as);
		assertTrue(cm.setAccountId(AC1_ID));
		assertFalse(cm.setPin(AC1_PIN + 1));
		assertFalse(cm.setPin(AC1_PIN + 1));
		assertFalse(cm.setPin(AC1_PIN + 1));
		verify(as).banAccount(AC1_ID);
		org.mockito.InOrder aso = inOrder(as);
		aso.verify(as, times(1)).isValidAccountId(AC1_ID);
		aso.verify(as, times(3)).validatePin(AC1_ID, AC1_PIN + 1);
	}
}
