package cashmachine;

import static org.junit.Assert.*;

import java.util.Arrays;

import org.junit.Before;
import org.junit.Test;

import cashmachine.api.AccountService;
import cashmachine.api.CashMachineException;
import cashmachine.impl.SimpleAccountService;
import cashmachine.model.Account;

public class SimpleAccountServiceTest
{
	private static final int AC1_ID = 1;
	private static final int AC1_PIN = 11;
	private static final int AC1_BALANCE = 1000;
	private static final int AC1_LIMIT = 100;
	private static final int AC2_ID = 2;
	private static final int AC2_PIN = 22;
	private static final int AC2_BALANCE = 1000;
	private static final int AC2_LIMIT = 1000;
	private static final int INVALID_ID = 3;

	private AccountService as;

	@Before
	public void setup() {
		Account ac1 = new Account(AC1_ID, AC1_PIN, AC1_BALANCE, AC1_LIMIT, Arrays.asList(AC1_ID, AC2_ID));
		Account ac2 = new Account(AC2_ID, AC2_PIN, AC2_BALANCE, AC2_LIMIT, Arrays.asList(AC2_ID));
		as = new SimpleAccountService(Arrays.asList(ac1, ac2));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void isValidAccountId_shouldBeValid()
	{
		assertTrue(as.isValidAccountId(AC1_ID));
		assertTrue(as.isValidAccountId(AC2_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void isValidAccountId_InvalidId()
	{
		assertFalse(as.isValidAccountId(INVALID_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void banAccount_shouldGetBanned()
	{
		/* ban account AC1_ID */
		assertTrue(as.isValidAccountId(AC1_ID));
		assertTrue(as.banAccount(AC1_ID));
		assertTrue(as.banAccount(AC1_ID));
		assertFalse(as.isValidAccountId(AC1_ID));

		/* ban account AC2_ID */
		assertTrue(as.isValidAccountId(AC2_ID));
		assertTrue(as.banAccount(AC2_ID));
		assertTrue(as.banAccount(AC2_ID));
		assertFalse(as.isValidAccountId(AC2_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void banAccount_InvalidAccount()
	{
		assertFalse(as.banAccount(INVALID_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void validatePin_shouldBeValid()
	{
		/* get AC1_ID */
		Account ac1 = as.validatePin(AC1_ID, AC1_PIN);
		assertNotNull(ac1);
		assertEquals(AC1_ID,      ac1.getAccountId());
		assertEquals(0,           ac1.getPin());
		assertEquals(AC1_BALANCE, ac1.getBalance());
		assertEquals(AC1_LIMIT,   ac1.getLimit());

		/* get AC2_ID */
		Account ac2 = as.validatePin(AC2_ID, AC2_PIN);
		assertNotNull(ac2);
		assertEquals(AC2_ID,      ac2.getAccountId());
		assertEquals(0,           ac2.getPin());
		assertEquals(AC2_BALANCE, ac2.getBalance());
		assertEquals(AC2_LIMIT,   ac2.getLimit());
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void validatePin_InvalidPin()
	{
		assertNull(as.validatePin(AC1_ID, AC1_PIN + 1));
		assertNull(as.validatePin(AC2_ID, AC2_PIN + 1));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void validatePin_InvalidAccount()
	{
		assertNull(as.validatePin(INVALID_ID, 0));
	}
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void getWithdrawalAccount_shouldBeValid()
	{
		/* get AC1_ID using AC1_ID */
		Account ac1 = as.getWithdrawalAccount(AC1_ID, AC1_PIN, AC1_ID);
		assertNotNull(ac1);
		assertEquals(AC1_ID,      ac1.getAccountId());
		assertEquals(0,           ac1.getPin());
		assertEquals(AC1_BALANCE, ac1.getBalance());
		assertEquals(AC1_LIMIT,   ac1.getLimit());

		/* get AC2_ID using AC1_ID */
		Account ac2 = as.getWithdrawalAccount(AC1_ID, AC1_PIN, AC2_ID);
		assertNotNull(ac2);
		assertEquals(AC2_ID,      ac2.getAccountId());
		assertEquals(0,           ac2.getPin());
		assertEquals(AC2_BALANCE, ac2.getBalance());
		assertEquals(AC2_LIMIT,   ac2.getLimit());

		/* get AC2_ID using AC2_ID */
		Account ac3 = as.getWithdrawalAccount(AC2_ID, AC2_PIN, AC2_ID);
		assertNotNull(ac3);
		assertEquals(AC2_ID,      ac3.getAccountId());
		assertEquals(0,           ac3.getPin());
		assertEquals(AC2_BALANCE, ac3.getBalance());
		assertEquals(AC2_LIMIT,   ac3.getLimit());
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void getWithdrawalAccount_invalidPin()
	{
		assertNull(as.getWithdrawalAccount(AC1_ID, AC1_PIN + 1, AC1_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void getWithdrawalAccount_invalidAccount()
	{
		assertNull(as.getWithdrawalAccount(AC1_ID, AC1_PIN, INVALID_ID));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test
	public void withdraw_shouldBeValid()
	throws CashMachineException
	{
		assertTrue(as.withdraw(AC1_ID, AC1_PIN, AC1_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE)) > 0);
		assertTrue(as.withdraw(AC1_ID, AC1_PIN, AC2_ID,
				java.lang.Math.min(AC2_LIMIT, AC2_BALANCE)) > 0);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void withdraw_invalidAccount()
	throws CashMachineException
	{
		as.withdraw(INVALID_ID, AC1_PIN, AC1_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void withdraw_invalidPin()
	throws CashMachineException
	{
		as.withdraw(AC1_ID, AC1_PIN + 1, AC1_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void withdraw_invalidWithdrawalAccount()
	throws CashMachineException
	{
		as.withdraw(AC1_ID, AC1_PIN, INVALID_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void withdraw_invalidAmount()
	throws CashMachineException
	{
		as.withdraw(AC1_ID, AC1_PIN, AC1_ID, 0);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void withdraw_amountTooHigh()
	throws CashMachineException
	{
		assertEquals(0, as.withdraw(AC1_ID, AC1_PIN, AC1_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE) + 1));
		assertEquals(0, as.withdraw(AC1_ID, AC1_PIN, AC2_ID,
				java.lang.Math.min(AC2_LIMIT, AC2_BALANCE) + 1));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void revertWithdraw_shouldBeValid()
	throws CashMachineException
	{
		/*
		 * amount = min(AC2_LIMIT, AC2_BALANCE)
		 * withdraw amount - keep from AC2_ID using AC2_ID
		 * withdraw keep from AC2_ID using AC1_ID
		 * revert both withdraws
		 * withdraw amount from AC2_ID using AC2_ID
		 */
		int amount = java.lang.Math.min(AC2_LIMIT, AC2_BALANCE);
		int keep = amount / 2;
		int transid1 = as.withdraw(AC2_ID, AC2_PIN, AC2_ID, amount - keep);
		assertTrue(transid1 > 0);
		int transid2 = as.withdraw(AC1_ID, AC1_PIN, AC2_ID, keep);
		assertTrue(transid2 > 0);
		assertTrue(as.revertWithdraw(AC2_ID, AC2_PIN, transid1));
		assertTrue(as.revertWithdraw(AC1_ID, AC1_PIN, transid2));
		assertTrue(as.withdraw(AC2_ID, AC2_PIN, AC2_ID, amount) > 0);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void revertWithdraw_invalidAccount()
	throws CashMachineException
	{
		as.revertWithdraw(INVALID_ID, 0, 0);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test(expected = CashMachineException.class)
	public void revertWithdraw_invalidPin()
	throws CashMachineException
	{
		as.revertWithdraw(AC1_ID, AC1_PIN + 1, 0);
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void revertWithdraw_invalidTransid()
	throws CashMachineException
	{
		assertFalse(as.revertWithdraw(AC1_ID, AC1_PIN, 0));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void revertWithdraw_useTransidTwice()
	throws CashMachineException
	{
		int transid = as.withdraw(AC1_ID, AC1_PIN, AC1_ID,
				java.lang.Math.min(AC1_LIMIT, AC1_BALANCE));
		assertTrue(transid > 0);
		assertTrue(as.revertWithdraw(AC1_ID, AC1_PIN, transid));
		assertFalse(as.revertWithdraw(AC1_ID, AC1_PIN, transid));
	}
	
	/*--------------------------------------------------------------------------*/
	
	@Test()
	public void revertWithdraw_usWrongTransids()
	throws CashMachineException
	{
		int amount = java.lang.Math.min(AC2_LIMIT, AC2_BALANCE);
		int keep = amount / 2;
		int transid1 = as.withdraw(AC1_ID, AC1_PIN, AC2_ID, amount - keep);
		assertTrue(transid1 > 0);
		int transid2 = as.withdraw(AC2_ID, AC2_PIN, AC2_ID, keep);
		assertTrue(transid2 > 0);
		assertFalse(as.revertWithdraw(AC1_ID, AC1_PIN, transid2));
		assertFalse(as.revertWithdraw(AC2_ID, AC2_PIN, transid1));
	}
}
